from weakref import WeakKeyDictionary
import pygame
from Vector import Vector, angle_vector
import Geometry
import Image
import PhysicsController
#from data.Rope import Rope, BounceRope, ExactRope
import data.Rope as Rope
from data.Object import Object, Powerup, Player, Enemy
import Level
from Level import GameOverException
import Event
import ode

class EndOfObjectException(Exception):
    def __init__(self, obj):
        self.obj = obj

m = ode.Mass()
m.setSphere(1, 16) # mass params for player
class Game(object):
    FORCESCALE = m.mass
    def __init__(self, evman):
        #self.player = Player(Vector([200, 200]))
        self.geometry = Geometry.Geometry()

        self.bullets = []
        self.explosions = []
        self.delete_next = []
        self.ropes = {}
        self.goals = []
        self.patterns = WeakKeyDictionary()
        self.enemies = []
        self.frames = 0
        self.forces = {}
        self.impulses = {}
        self.actions = []
        self.flags = {} # used for setting game state
        self.thrustpos = None
        self.evmanager = evman
        self.fps = 0
        self.paused = False

    def kill_all(self):
        self.patterns = WeakKeyDictionary()
        self.evmanager = None

    def delete_object(self, object, killer=None, chrome=True):
        self.post(Event.ObjectDestroyEvent(object, killer, chrome=chrome))
        self.geometry.delete_object(object)
        if object in self.patterns.keys():
            del self.patterns[object]
        if hasattr(object,'on_destroy'):
            #print 'setting pattern for', object
            self.patterns[object] = object.on_destroy
        if object in self.forces.keys():
            del self.forces[object]
        if object in self.impulses.keys():
            del self.impulses[object]
        for lst in self.bullets, self.enemies:
            if object in lst:
                lst.remove(object)
        for player, rope in self.ropes.items():
            if object in rope.objects:
                del self.ropes[player]

    def delete_part(self, object, label):
        g = self.geometry.geoms[object][label]
        self.post(Event.ObjectDestroyPartEvent(g))
        self.geometry.delete_part(object, label)
        

    def queue_delete(self, object, killstr=None):
        '''It is sometimes helpful to wait to delete an object.'''
        self.delete_next.append([object, killstr])
    delete = queue_delete

    def flush_delete(self):
        deleted = []
        for [o, killstr] in self.delete_next:
            if o not in deleted:
                self.delete_object(o, killstr)
                deleted.append(o)
        self.delete_next = []
        
    def objects(self):
        return self.geometry.objects()

    def shoot(self, shooter, bullet):
        self.bullets.append(bullet)
        p = self.geometry.get_pos(shooter)
        bulletpos = shooter.bulletpos
        if not bulletpos:
            bulletpos = p
            angle = shooter.angle
            bulletpos = bulletpos + angle_vector(angle)*(shooter.size+bullet.size+.001)
        pos = bulletpos
        speed = 10
        if shooter.bulletspeed:
            speed = shooter.bulletspeed
        self.geometry.new_body(bullet, pos, speed*angle_vector(shooter.angle)+self.geometry.get_velo(shooter), (0, 0))
##        print 'bulletvelo',10*angle_vector(shooter.angle)
        if hasattr(bullet, 'imagename'): bulletimage = bullet.imagename
        else: bulletimage = shooter.bulletimage
        image = Image.find_image(bulletimage, self.currentfile)
        #s = [pygame.transform.rotate(i, shooter.angle) for i in image]
        s = [i for i in image]
        self.geometry.add_circle(bullet, 0, bullet.size, density=1, image=Image.Strip(s))
        bullet.add_geometry(self.geometry)
        self.post(Event.ObjectRefreshEvent(bullet))
        self.post(Event.ObjectShootingEvent(shooter))

        self.play_sound(shooter.bulletsound)
        return bullet

    def make_explosion(self, center, angle, pic, sound=None):
        self.post(Event.ExplosionSpriteEvent(center, angle, pic))
        if sound:
            self.post(Event.NewSoundEvent(sound))
    def skip_dialog(self):
        for d in self.goals:
            if isinstance(d, Level.Dialog):
                d.skip()
                return True

    def pause(self, start):
        self.paused = start
                
    def hit(self, thing, killstr, strength):
        if thing.dead: return # who cares?
        if strength == None: # destroy!!!
            thing.health = 0
        elif not isinstance(thing.health, int):
            return # indestructable
        else: thing.health -= strength

        self.post(Event.ObjectHitEvent(thing, killstr, thing.health <= 0))
        if thing.health <= 0:
            # thing dies!!
            thing.die(self)
            if thing != self.player:
                self.queue_delete(thing, killstr)
            else:
                self.post(Event.ObjectDestroyEvent(thing, killstr, True))
                pass

            if thing.diesound:
                self.post(Event.NewSoundEvent(thing.diesound))


    def add_rope(self, ropetype, player, o2, pos, d):
        r=ropetype(self, player, o2, pos, d)
        self.ropes[player]= r

    def unhook(self, player):
        del self.ropes[player]

    def hooked(self, player):
        return player in self.ropes.keys()

    def make_inanim(self, obj):
        self.post(Event.ObjectIncomingEvent(obj))

    def add_enemy(self, e):
        self.enemies.append(e)

    def add_image(self, *args, **kwargs):
        e = Event.AddImageEvent(*args, **kwargs)
        self.post(e)
        return e.image

    def thread(self):
        while True:
            #print self.player, 'velo', self.geometry.get_velo(self.player)
            #print 'bullets', self.bullets
            if not self.paused:
                self.flush_delete()
                self.frames += 1
            yield None

    def add_force(self, object, force):
        self.forces[object] = self.forces.get(object, Vector([0, 0])) + force

    def add_impulse(self, object, force):
        self.impulses[object] = self.forces.get(object, Vector([0, 0])) + force

    def add_action(self, action):
        self.actions.append(action)

    def post(self, event):
        self.evmanager.Post(event)

    def play_music(self, mfile, loop):
        self.evmanager.Post(Event.NewSoundEvent(mfile, loop, music=True))

    def play_sound(self, sfile):
        self.evmanager.Post(Event.NewSoundEvent(sfile))
        
